@@ -3,7 +3,7 @@
  *
  * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
  * Julian Hall. All rights reserved. The software is
- * redistributable under the licence given in the file "Licence"
+ * redistributable under the license given in the file "LICENSE"
  * distributed in the NASM archive.
  */
 
@@ -17,6 +17,8 @@
 
 #include "nasm.h"
 #include "nasmlib.h"
+#include "saa.h"
+#include "raa.h"
 #include "stdscan.h"
 #include "outform.h"
 
@@ -110,6 +112,9 @@ static struct Symbol *fwds;
 
 static char elf_module[FILENAME_MAX];
 
+static uint8_t elf_osabi = 0;	/* Default OSABI = 0 (System V or Linux) */
+static uint8_t elf_abiver = 0;	/* Current ABI version */
+
 extern struct ofmt of_elf32;
 extern struct ofmt of_elf;
 
@@ -119,6 +124,8 @@ extern struct ofmt of_elf;
 
 #define SYM_GLOBAL 0x10
 
+#define SHT_RELA	  4		/* Relocation entries with addends */
+
 #define STT_NOTYPE	0		/* Symbol type is unspecified */
 #define STT_OBJECT	1		/* Symbol is a data object */
 #define STT_FUNC	2		/* Symbol is a code object */
@@ -133,20 +140,44 @@ extern struct ofmt of_elf;
 #define STV_HIDDEN 2
 #define STV_PROTECTED 3
 
-#define GLOBAL_TEMP_BASE 16     /* bigger than any constant sym id */
+#define GLOBAL_TEMP_BASE 1048576     /* bigger than any reasonable sym id */
 
 #define SEG_ALIGN 16            /* alignment of sections in file */
 #define SEG_ALIGN_1 (SEG_ALIGN-1)
 
+/* Definitions in lieu of dwarf.h */
+#define    DW_TAG_compile_unit   0x11
+#define    DW_TAG_subprogram   0x2e
+#define    DW_AT_name   0x03
+#define    DW_AT_stmt_list   0x10
+#define    DW_AT_low_pc   0x11
+#define    DW_AT_high_pc   0x12
+#define    DW_AT_language  0x13
+#define    DW_AT_producer   0x25
+#define    DW_AT_frame_base   0x40
+#define    DW_FORM_addr   0x01
+#define    DW_FORM_data2   0x05
+#define    DW_FORM_data4   0x06
+#define    DW_FORM_string   0x08
+#define    DW_LNS_extended_op  0
+#define    DW_LNS_advance_pc   2
+#define    DW_LNS_advance_line   3
+#define    DW_LNS_set_file   4
+#define    DW_LNE_end_sequence   1
+#define    DW_LNE_set_address   2
+#define    DW_LNE_define_file   3
+#define    DW_LANG_Mips_Assembler  0x8001
+
+#define SOC(ln,aa) ln - line_base + (line_range * aa) + opcode_base
+
 static const char align_str[SEG_ALIGN] = "";    /* ANSI will pad this with 0s */
 
-#define ELF_MAX_SECTIONS 16     /* really 10, but let's play safe */
 static struct ELF_SECTDATA {
     void *data;
     int32_t len;
     bool is_saa;
 } *elf_sects;
-static int elf_nsect;
+static int elf_nsect, nsections;
 static int32_t elf_foffs;
 
 static void elf_write(void);
@@ -193,28 +224,64 @@ struct linelist {
     struct linelist *last;
 };
 
+struct sectlist {
+    struct SAA *psaa;
+    int section;
+    int line;
+    int offset;
+    int file;
+    struct sectlist *next;
+    struct sectlist *last;
+};
+
+/* common debug variables */
+static int currentline = 1;
+static int debug_immcall = 0;
+
+/* stabs debug variables */
 static struct linelist *stabslines = 0;
-static int stabs_immcall = 0;
-static int currentline = 0;
 static int numlinestabs = 0;
 static char *stabs_filename = 0;
 static int symtabsection;
 static uint8_t *stabbuf = 0, *stabstrbuf = 0, *stabrelbuf = 0;
 static int stablen, stabstrlen, stabrellen;
 
+/* dwarf debug variables */
+static struct linelist *dwarf_flist = 0, *dwarf_clist = 0, *dwarf_elist = 0;
+static struct sectlist *dwarf_fsect = 0, *dwarf_csect = 0, *dwarf_esect = 0;
+static int dwarf_numfiles = 0, dwarf_nsections;
+static uint8_t *arangesbuf = 0, *arangesrelbuf = 0, *pubnamesbuf = 0, *infobuf = 0,  *inforelbuf = 0,
+               *abbrevbuf = 0, *linebuf = 0, *linerelbuf = 0, *framebuf = 0, *locbuf = 0;
+static int8_t line_base = -5, line_range = 14, opcode_base = 13;
+static int arangeslen, arangesrellen, pubnameslen, infolen, inforellen,
+           abbrevlen, linelen, linerellen, framelen, loclen;
+static int32_t dwarf_infosym, dwarf_abbrevsym, dwarf_linesym;
+
+static struct dfmt df_dwarf;
 static struct dfmt df_stabs;
 static struct Symbol *lastsym;
 
-void stabs32_init(struct ofmt *, void *, FILE *, efunc);
+/* common debugging routines */
+void debug32_typevalue(int32_t);
+void debug32_init(struct ofmt *, void *, FILE *, efunc);
+void debug32_deflabel(char *, int32_t, int64_t, int, char *);
+void debug32_directive(const char *, const char *);
+
+/* stabs debugging routines */
 void stabs32_linenum(const char *filename, int32_t linenumber, int32_t);
-void stabs32_deflabel(char *, int32_t, int32_t, int, char *);
-void stabs32_directive(const char *, const char *);
-void stabs32_typevalue(int32_t);
 void stabs32_output(int, void *);
 void stabs32_generate(void);
 void stabs32_cleanup(void);
 
-/* end of stabs debugging stuff */
+/* dwarf debugging routines */
+void dwarf32_linenum(const char *filename, int32_t linenumber, int32_t);
+void dwarf32_output(int, void *);
+void dwarf32_generate(void);
+void dwarf32_cleanup(void);
+void dwarf32_findfile(const char *);
+void dwarf32_findsect(const int);
+void saa_wleb128u(struct SAA *, int);
+void saa_wleb128s(struct SAA *, int);
 
 /*
  * Special section numbers which are used to define ELF special
@@ -227,6 +294,8 @@ static int32_t elf_sym_sect;
 
 static void elf_init(FILE * fp, efunc errfunc, ldfunc ldef, evalfunc eval)
 {
+    if (of_elf.current_dfmt != &null_debug_form)
+        of_elf32.current_dfmt = of_elf.current_dfmt;
     elffp = fp;
     error = errfunc;
     evaluate = eval;
@@ -238,7 +307,7 @@ static void elf_init(FILE * fp, efunc errfunc, ldfunc ldef, evalfunc eval)
     bsym = raa_init();
     strs = saa_init(1L);
     saa_wbytes(strs, "\0", 1L);
-    saa_wbytes(strs, elf_module, (int32_t)(strlen(elf_module) + 1));
+    saa_wbytes(strs, elf_module, strlen(elf_module)+1);
     strslen = 2 + strlen(elf_module);
     shstrtab = NULL;
     shstrtablen = shstrtabsize = 0;;
@@ -351,21 +420,21 @@ static int32_t elf_section_names(char *name, int pass, int *bits)
     }
 
     p = name;
-    while (*p && !isspace(*p))
+    while (*p && !nasm_isspace(*p))
         p++;
     if (*p)
         *p++ = '\0';
     flags_and = flags_or = type = align = 0;
 
-    while (*p && isspace(*p))
+    while (*p && nasm_isspace(*p))
         p++;
     while (*p) {
         char *q = p;
-        while (*p && !isspace(*p))
+        while (*p && !nasm_isspace(*p))
             p++;
         if (*p)
             *p++ = '\0';
-        while (*p && isspace(*p))
+        while (*p && nasm_isspace(*p))
             p++;
 
         if (!nasm_strnicmp(q, "align=", 6)) {
@@ -444,7 +513,7 @@ static int32_t elf_section_names(char *name, int pass, int *bits)
     return sects[i]->index;
 }
 
-static void elf_deflabel(char *name, int32_t segment, int32_t offset,
+static void elf_deflabel(char *name, int32_t segment, int64_t offset,
                          int is_global, char *special)
 {
     int pos = strslen;
@@ -481,9 +550,9 @@ static void elf_deflabel(char *name, int32_t segment, int32_t offset,
                 expr *e;
                 char *p = special;
 
-                while (*p && !isspace(*p))
+                while (*p && !nasm_isspace(*p))
                     p++;
-                while (*p && isspace(*p))
+                while (*p && nasm_isspace(*p))
                     p++;
                 stdscan_reset();
                 stdscan_bufptr = p;
@@ -600,7 +669,7 @@ static void elf_deflabel(char *name, int32_t segment, int32_t offset,
                           n, special);
                 special += n;
 
-                while (isspace(*special))
+                while (nasm_isspace(*special))
                     ++special;
                 if (*special) {
                     n = strcspn(special, " \t");
@@ -623,7 +692,7 @@ static void elf_deflabel(char *name, int32_t segment, int32_t offset,
                     int fwd = 0;
                     char *saveme = stdscan_bufptr;      /* bugfix? fbk 8/10/00 */
 
-                    while (special[n] && isspace(special[n]))
+                    while (special[n] && nasm_isspace(special[n]))
                         n++;
                     /*
                      * We have a size expression; attempt to
@@ -772,18 +841,16 @@ static int32_t elf_add_gsym_reloc(struct Section *sect,
     return offset - sym->value;
 }
 
-static void elf_out(int32_t segto, const void *data, uint32_t type,
+static void elf_out(int32_t segto, const void *data,
+		    enum out_type type, uint64_t size,
                     int32_t segment, int32_t wrt)
 {
     struct Section *s;
-    int32_t realbytes = type & OUT_SIZMASK;
     int32_t addr;
     uint8_t mydata[4], *p;
     int i;
     static struct symlininfo sinfo;
 
-    type &= OUT_TYPMASK;
-
     /*
      * handle absolute-assembly (structure definitions)
      */
@@ -823,10 +890,10 @@ static void elf_out(int32_t segto, const void *data, uint32_t type,
         error(ERR_WARNING, "attempt to initialize memory in"
               " BSS section `%s': ignored", s->name);
         if (type == OUT_REL2ADR)
-            realbytes = 2;
+            size = 2;
         else if (type == OUT_REL4ADR)
-            realbytes = 4;
-        s->len += realbytes;
+            size = 4;
+        s->len += size;
         return;
     }
 
@@ -834,23 +901,23 @@ static void elf_out(int32_t segto, const void *data, uint32_t type,
         if (s->type == SHT_PROGBITS) {
             error(ERR_WARNING, "uninitialized space declared in"
                   " non-BSS section `%s': zeroing", s->name);
-            elf_sect_write(s, NULL, realbytes);
+            elf_sect_write(s, NULL, size);
         } else
-            s->len += realbytes;
+            s->len += size;
     } else if (type == OUT_RAWDATA) {
         if (segment != NO_SEG)
             error(ERR_PANIC, "OUT_RAWDATA with other than NO_SEG");
-        elf_sect_write(s, data, realbytes);
+        elf_sect_write(s, data, size);
     } else if (type == OUT_ADDRESS) {
         bool gnu16 = false;
-        addr = *(int32_t *)data;
+        addr = *(int64_t *)data;
         if (segment != NO_SEG) {
             if (segment % 2) {
                 error(ERR_NONFATAL, "ELF format does not support"
                       " segment base references");
             } else {
                 if (wrt == NO_SEG) {
-                    if (realbytes == 2) {
+                    if (size == 2) {
                         gnu16 = true;
                         elf_add_reloc(s, segment, R_386_16);
                     } else {
@@ -870,7 +937,7 @@ static void elf_out(int32_t segto, const void *data, uint32_t type,
                     addr = elf_add_gsym_reloc(s, segment, addr,
                                               R_386_GOT32, true);
                 } else if (wrt == elf_sym_sect + 1) {
-                    if (realbytes == 2) {
+                    if (size == 2) {
                         gnu16 = true;
                         addr = elf_add_gsym_reloc(s, segment, addr,
                                                   R_386_16, false);
@@ -894,13 +961,13 @@ static void elf_out(int32_t segto, const void *data, uint32_t type,
                   "16-bit relocations in ELF is a GNU extension");
             WRITESHORT(p, addr);
         } else {
-            if (realbytes != 4 && segment != NO_SEG) {
+            if (size != 4 && segment != NO_SEG) {
                 error(ERR_NONFATAL,
                       "Unsupported non-32-bit ELF relocation");
             }
             WRITELONG(p, addr);
         }
-        elf_sect_write(s, mydata, realbytes);
+        elf_sect_write(s, mydata, size);
     } else if (type == OUT_REL2ADR) {
         if (segment == segto)
             error(ERR_PANIC, "intra-segment OUT_REL2ADR");
@@ -918,7 +985,7 @@ static void elf_out(int32_t segto, const void *data, uint32_t type,
             }
         }
         p = mydata;
-        WRITESHORT(p, *(int32_t *)data - realbytes);
+        WRITESHORT(p, *(int64_t *)data - size);
         elf_sect_write(s, mydata, 2L);
     } else if (type == OUT_REL4ADR) {
         if (segment == segto)
@@ -943,14 +1010,14 @@ static void elf_out(int32_t segto, const void *data, uint32_t type,
             }
         }
         p = mydata;
-        WRITELONG(p, *(int32_t *)data - realbytes);
+        WRITELONG(p, *(int64_t *)data - size);
         elf_sect_write(s, mydata, 4L);
     }
 }
 
 static void elf_write(void)
 {
-    int nsections, align;
+    int align;
     int scount;
     char *p;
     int commlen;
@@ -968,6 +1035,8 @@ static void elf_write(void)
      */
     if (of_elf32.current_dfmt == &df_stabs)
         nsections = 8;
+    else if (of_elf32.current_dfmt == &df_dwarf)
+        nsections = 15;
     else
         nsections = 5;          /* SHN_UNDEF and the fixed ones */
 
@@ -990,6 +1059,26 @@ static void elf_write(void)
         add_sectname(".rel", ".stab");
     }
 
+    else if (of_elf32.current_dfmt == &df_dwarf) {
+        /* the dwarf debug standard specifies the following ten sections,
+           not all of which are currently implemented,
+           although all of them are defined. */
+        #define debug_aranges (int32_t) (nsections-10)
+        #define debug_info (int32_t) (nsections-7)
+        #define debug_abbrev (int32_t) (nsections-5)
+        #define debug_line (int32_t) (nsections-4)
+        add_sectname("", ".debug_aranges");
+        add_sectname(".rela", ".debug_aranges");
+        add_sectname("", ".debug_pubnames");
+        add_sectname("", ".debug_info");
+        add_sectname(".rela", ".debug_info");
+        add_sectname("", ".debug_abbrev");
+        add_sectname("", ".debug_line");
+        add_sectname(".rela", ".debug_line");
+        add_sectname("", ".debug_frame");
+        add_sectname("", ".debug_loc");
+    }
+
     /*
      * Do the comment.
      */
@@ -1000,7 +1089,10 @@ static void elf_write(void)
     /*
      * Output the ELF header.
      */
-    fwrite("\177ELF\1\1\1\0\0\0\0\0\0\0\0\0", 16, 1, elffp);
+    fwrite("\177ELF\1\1\1", 7, 1, elffp);
+    fputc(elf_osabi, elffp);
+    fputc(elf_abiver, elffp);
+    fwrite("\0\0\0\0\0\0\0", 7, 1, elffp);
     fwriteint16_t(1, elffp);      /* ET_REL relocatable file */
     fwriteint16_t(3, elffp);      /* EM_386 processor ID */
     fwriteint32_t(1L, elffp);      /* EV_CURRENT file format version */
@@ -1037,7 +1129,7 @@ static void elf_write(void)
     align = ((elf_foffs + SEG_ALIGN_1) & ~SEG_ALIGN_1) - elf_foffs;
     elf_foffs += align;
     elf_nsect = 0;
-    elf_sects = nasm_malloc(sizeof(*elf_sects) * (2 * nsects + 10));
+    elf_sects = nasm_malloc(sizeof(*elf_sects) * nsections);
 
     elf_section_header(0, 0, 0, NULL, false, 0L, 0, 0, 0, 0);   /* SHN_UNDEF */
     scount = 1;                 /* needed for the stabs debugging to track the symtable section */
@@ -1089,6 +1181,44 @@ static void elf_write(void)
                                8);
         }
     }
+    else if (of_elf32.current_dfmt == &df_dwarf) {
+            /* for dwarf debugging information, create the ten dwarf sections */
+
+            /* this function call creates the dwarf sections in memory */
+            if (dwarf_fsect) dwarf32_generate();
+
+            p += strlen(p) + 1;
+            elf_section_header(p - shstrtab, SHT_PROGBITS, 0, arangesbuf, false,
+                               arangeslen, 0, 0, 1, 0);
+            p += strlen(p) + 1;
+            elf_section_header(p - shstrtab, SHT_RELA, 0, arangesrelbuf, false,
+                               arangesrellen, symtabsection, debug_aranges, 1, 12);
+            p += strlen(p) + 1;
+            elf_section_header(p - shstrtab, SHT_PROGBITS, 0, pubnamesbuf, false,
+                               pubnameslen, 0, 0, 1, 0);
+            p += strlen(p) + 1;
+            elf_section_header(p - shstrtab, SHT_PROGBITS, 0, infobuf, false,
+                               infolen, 0, 0, 1, 0);
+            p += strlen(p) + 1;
+            elf_section_header(p - shstrtab, SHT_RELA, 0, inforelbuf, false,
+                               inforellen, symtabsection, debug_info, 1, 12);
+            p += strlen(p) + 1;
+            elf_section_header(p - shstrtab, SHT_PROGBITS, 0, abbrevbuf, false,
+                               abbrevlen, 0, 0, 1, 0);
+            p += strlen(p) + 1;
+            elf_section_header(p - shstrtab, SHT_PROGBITS, 0, linebuf, false,
+                               linelen, 0, 0, 1, 0);
+            p += strlen(p) + 1;
+            elf_section_header(p - shstrtab, SHT_RELA, 0, linerelbuf, false,
+                               linerellen, symtabsection, debug_line, 1, 12);
+            p += strlen(p) + 1;
+            elf_section_header(p - shstrtab, SHT_PROGBITS, 0, framebuf, false,
+                               framelen, 0, 0, 8, 0);
+            p += strlen(p) + 1;
+            elf_section_header(p - shstrtab, SHT_PROGBITS, 0, locbuf, false,
+                               loclen, 0, 0, 1, 0);
+
+    }
     fwrite(align_str, align, 1, elffp);
 
     /*
@@ -1163,6 +1293,43 @@ static struct SAA *elf_build_symtab(int32_t *len, int32_t *local)
         *len += 16;
         (*local)++;
     }
+     /*
+      * dwarf needs symbols for debug sections
+      * which are relocation targets.
+      */  
+//*** fix for 32 bit
+     if (of_elf32.current_dfmt == &df_dwarf) {
+        dwarf_infosym = *local;
+        p = entry;
+        WRITELONG(p, 0);        /* no symbol name */
+        WRITELONG(p, (uint32_t) 0);        /* offset zero */
+        WRITELONG(p, (uint32_t) 0);        /* size zero */
+        WRITESHORT(p, STT_SECTION);       /* type, binding, and visibility */
+        WRITESHORT(p, debug_info);       /* section id */
+        saa_wbytes(s, entry, 16L);
+        *len += 16;
+        (*local)++;
+        dwarf_abbrevsym = *local;
+        p = entry;
+        WRITELONG(p, 0);        /* no symbol name */
+        WRITELONG(p, (uint32_t) 0);        /* offset zero */
+        WRITELONG(p, (uint32_t) 0);        /* size zero */
+        WRITESHORT(p, STT_SECTION);       /* type, binding, and visibility */
+        WRITESHORT(p, debug_abbrev);       /* section id */
+        saa_wbytes(s, entry, 16L);
+        *len += 16;
+        (*local)++;
+        dwarf_linesym = *local;
+        p = entry;
+        WRITELONG(p, 0);        /* no symbol name */
+        WRITELONG(p, (uint32_t) 0);        /* offset zero */
+        WRITELONG(p, (uint32_t) 0);        /* size zero */
+        WRITESHORT(p, STT_SECTION);       /* type, binding, and visibility */
+        WRITESHORT(p, debug_line);       /* section id */
+        saa_wbytes(s, entry, 16L);
+        *len += 16;
+        (*local)++;
+     }
 
     /*
      * Now the global symbols.
@@ -1200,7 +1367,11 @@ static struct SAA *elf_build_reltab(int32_t *len, struct Reloc *r)
         int32_t sym = r->symbol;
 
         if (sym >= GLOBAL_TEMP_BASE)
-            sym += -GLOBAL_TEMP_BASE + (nsects + 2) + nlocals;
+        {
+           if (of_elf32.current_dfmt == &df_dwarf)
+              sym += -GLOBAL_TEMP_BASE + (nsects + 5) + nlocals;
+           else   sym += -GLOBAL_TEMP_BASE + (nsects + 2) + nlocals;
+        }
 
         p = entry;
         WRITELONG(p, r->address);
@@ -1267,9 +1438,39 @@ static int32_t elf_segbase(int32_t segment)
 
 static int elf_directive(char *directive, char *value, int pass)
 {
-    (void)directive;
-    (void)value;
-    (void)pass;
+    bool err;
+    int64_t n;
+    char *p;
+
+    if (!strcmp(directive, "osabi")) {
+	if (pass == 2)
+	    return 1;		/* ignore in pass 2 */
+
+	n = readnum(value, &err);
+	if (err) {
+	    error(ERR_NONFATAL, "`osabi' directive requires a parameter");
+	    return 1;
+	}
+	if (n < 0 || n > 255) {
+	    error(ERR_NONFATAL, "valid osabi numbers are 0 to 255");
+	    return 1;
+	}
+	elf_osabi  = n;
+	elf_abiver = 0;
+
+	if ((p = strchr(value,',')) == NULL)
+	    return 1;
+
+	n = readnum(p+1, &err);
+	if (err || n < 0 || n > 255) {
+	    error(ERR_NONFATAL, "invalid ABI version number (valid: 0 to 255)");
+	    return 1;
+	}
+	
+	elf_abiver = n;
+	return 1;
+    }
+	
     return 0;
 }
 
@@ -1279,33 +1480,38 @@ static void elf_filename(char *inname, char *outname, efunc error)
     standard_extension(inname, outname, ".o", error);
 }
 
-static const char *elf_stdmac[] = {
-    "%define __SECT__ [section .text]",
-    "%macro __NASM_CDecl__ 1",
-    "%define $_%1 $%1",
-    "%endmacro",
-    NULL
-};
+extern macros_t elf_stdmac[];
+
 static int elf_set_info(enum geninfo type, char **val)
 {
     (void)type;
     (void)val;
     return 0;
 }
-
+static struct dfmt df_dwarf = {
+    "ELF32 (i386) dwarf debug format for Linux",
+    "dwarf",
+    debug32_init,
+    dwarf32_linenum,
+    debug32_deflabel,
+    debug32_directive,
+    debug32_typevalue,
+    dwarf32_output,
+    dwarf32_cleanup
+};
 static struct dfmt df_stabs = {
     "ELF32 (i386) stabs debug format for Linux",
     "stabs",
-    stabs32_init,
+    debug32_init,
     stabs32_linenum,
-    stabs32_deflabel,
-    stabs32_directive,
-    stabs32_typevalue,
+    debug32_deflabel,
+    debug32_directive,
+    debug32_typevalue,
     stabs32_output,
     stabs32_cleanup
 };
 
-struct dfmt *elf32_debugs_arr[2] = { &df_stabs, NULL };
+struct dfmt *elf32_debugs_arr[3] = { &df_stabs, &df_dwarf, NULL };
 
 struct ofmt of_elf32 = {
     "ELF32 (i386) object files (e.g. Linux)",
@@ -1344,7 +1550,7 @@ struct ofmt of_elf = {
 };
 /* again, the stabs debugging stuff (code) */
 
-void stabs32_init(struct ofmt *of, void *id, FILE * fp, efunc error)
+void debug32_init(struct ofmt *of, void *id, FILE * fp, efunc error)
 {
     (void)of;
     (void)id;
@@ -1371,11 +1577,11 @@ void stabs32_linenum(const char *filename, int32_t linenumber, int32_t segto)
             strcpy(stabs_filename, filename);
         }
     }
-    stabs_immcall = 1;
+    debug_immcall = 1;
     currentline = linenumber;
 }
 
-void stabs32_deflabel(char *name, int32_t segment, int32_t offset, int is_global,
+void debug32_deflabel(char *name, int32_t segment, int64_t offset, int is_global,
                     char *special)
 {
    (void)name;
@@ -1385,13 +1591,13 @@ void stabs32_deflabel(char *name, int32_t segment, int32_t offset, int is_global
    (void)special;
 }
 
-void stabs32_directive(const char *directive, const char *params)
+void debug32_directive(const char *directive, const char *params)
 {
    (void)directive;
    (void)params;
 }
 
-void stabs32_typevalue(int32_t type)
+void debug32_typevalue(int32_t type)
 {
     int32_t stype, ssize;
     switch (TYM_TYPE(type)) {
@@ -1448,7 +1654,7 @@ void stabs32_typevalue(int32_t type)
             stype = STT_NOTYPE;
             break;
     }
-    if (stype == STT_OBJECT && !lastsym->type) {
+    if (stype == STT_OBJECT && lastsym && !lastsym->type) {
         lastsym->size = ssize;
         lastsym->type = stype;
     }
@@ -1459,7 +1665,7 @@ void stabs32_output(int type, void *param)
     struct symlininfo *s;
     struct linelist *el;
     if (type == TY_STABSSYMLIN) {
-        if (stabs_immcall) {
+        if (debug_immcall) {
             s = (struct symlininfo *)param;
             if (!(sects[s->section]->flags & SHF_EXECINSTR))
                 return;         /* we are only interested in the text stuff */
@@ -1480,7 +1686,7 @@ void stabs32_output(int type, void *param)
             }
         }
     }
-    stabs_immcall = 0;
+    debug_immcall = 0;
 }
 
 #define WRITE_STAB(p,n_strx,n_type,n_other,n_desc,n_value) \
@@ -1643,5 +1849,440 @@ void stabs32_cleanup(void)
     if (stabstrbuf)
         nasm_free(stabstrbuf);
 }
+/* dwarf routines */
+
+
+void dwarf32_linenum(const char *filename, int32_t linenumber, int32_t segto)
+{
+    (void)segto;
+    dwarf32_findfile(filename);
+    debug_immcall = 1;
+    currentline = linenumber;
+}
+
+/* called from elf_out with type == TY_DEBUGSYMLIN */
+void dwarf32_output(int type, void *param)
+{
+  int ln, aa, inx, maxln, soc;
+  struct symlininfo *s;
+  struct SAA *plinep;
+
+  (void)type;
+
+  s = (struct symlininfo *)param;
+   /* line number info is only gathered for executable sections */
+   if (!(sects[s->section]->flags & SHF_EXECINSTR))
+     return;
+  /* Check if section index has changed */
+  if (!(dwarf_csect && (dwarf_csect->section) == (s->section)))
+  {
+     dwarf32_findsect(s->section);
+  }
+  /* do nothing unless line or file has changed */
+  if (debug_immcall)
+  {
+    ln = currentline - dwarf_csect->line;
+    aa = s->offset - dwarf_csect->offset;
+    inx = dwarf_clist->line;
+    plinep = dwarf_csect->psaa;
+    /* check for file change */
+    if (!(inx == dwarf_csect->file))
+    {
+       saa_write8(plinep,DW_LNS_set_file);
+       saa_write8(plinep,inx);
+       dwarf_csect->file = inx;
+    }
+    /* check for line change */
+    if (ln)
+    {
+       /* test if in range of special op code */
+       maxln = line_base + line_range;
+       soc = (ln - line_base) + (line_range * aa) + opcode_base;
+       if (ln >= line_base && ln < maxln && soc < 256)
+       {
+          saa_write8(plinep,soc);
+       }
+       else
+       {
+          if (ln)
+          {
+          saa_write8(plinep,DW_LNS_advance_line);
+          saa_wleb128s(plinep,ln);
+          }
+          if (aa)
+          {
+          saa_write8(plinep,DW_LNS_advance_pc);
+          saa_wleb128u(plinep,aa);
+          }
+       }
+       dwarf_csect->line = currentline;
+       dwarf_csect->offset = s->offset;
+    }
+    /* show change handled */
+    debug_immcall = 0;
+  }
+}
+
+
+void dwarf32_generate(void)
+{
+    static const char nasm_signature[] = "NASM " NASM_VER;
+    uint8_t *pbuf;
+    int indx;
+    struct linelist *ftentry;
+    struct SAA *paranges, *ppubnames, *pinfo, *pabbrev, *plines, *plinep;
+    struct SAA *parangesrel, *plinesrel, *pinforel;
+    struct sectlist *psect;
+    size_t saalen, linepoff, totlen, highaddr;
+
+    /* write epilogues for each line program range */
+    /* and build aranges section */
+    paranges = saa_init(1L);
+    parangesrel = saa_init(1L);
+    saa_write16(paranges,2);		/* dwarf version */
+    saa_write32(parangesrel, paranges->datalen+4);
+    saa_write32(parangesrel, (dwarf_infosym << 8) +  R_386_32); /* reloc to info */
+    saa_write32(parangesrel, 0);
+    saa_write32(paranges,0);		/* offset into info */
+    saa_write8(paranges,4);		/* pointer size */
+    saa_write8(paranges,0);		/* not segmented */
+    saa_write32(paranges,0);		/* padding */
+    /* iterate though sectlist entries */
+     psect = dwarf_fsect;
+     totlen = 0;
+     highaddr = 0;
+     for (indx = 0; indx < dwarf_nsections; indx++)
+     {
+         plinep = psect->psaa;
+         /* Line Number Program Epilogue */
+         saa_write8(plinep,2);			/* std op 2 */
+         saa_write8(plinep,(sects[psect->section]->len)-psect->offset);
+         saa_write8(plinep,DW_LNS_extended_op);
+         saa_write8(plinep,1);			/* operand length */
+         saa_write8(plinep,DW_LNE_end_sequence);
+         totlen += plinep->datalen;
+         /* range table relocation entry */
+         saa_write32(parangesrel, paranges->datalen + 4);
+         saa_write32(parangesrel, ((uint32_t) (psect->section + 2) << 8) +  R_386_32);
+         saa_write32(parangesrel, (uint32_t) 0);
+         /* range table entry */
+         saa_write32(paranges,0x0000);		/* range start */
+         saa_write32(paranges,sects[psect->section]->len);	/* range length */
+         highaddr += sects[psect->section]->len;
+         /* done with this entry */
+         psect = psect->next;
+     }
+    saa_write32(paranges,0);		/* null address */
+    saa_write32(paranges,0);		/* null length */
+    saalen = paranges->datalen;
+    arangeslen = saalen + 4;
+    arangesbuf = pbuf = nasm_malloc(arangeslen);
+    WRITELONG(pbuf,saalen);			/* initial length */
+    saa_rnbytes(paranges, pbuf, saalen);
+    saa_free(paranges);
+
+    /* build rela.aranges section */
+    arangesrellen = saalen = parangesrel->datalen;
+    arangesrelbuf = pbuf = nasm_malloc(arangesrellen); 
+    saa_rnbytes(parangesrel, pbuf, saalen);
+    saa_free(parangesrel);
+
+    /* build pubnames section */
+    ppubnames = saa_init(1L);
+    saa_write16(ppubnames,3);			/* dwarf version */
+    saa_write32(ppubnames,0);			/* offset into info */
+    saa_write32(ppubnames,0);			/* space used in info */
+    saa_write32(ppubnames,0);			/* end of list */
+    saalen = ppubnames->datalen;
+    pubnameslen = saalen + 4;
+    pubnamesbuf = pbuf = nasm_malloc(pubnameslen);
+    WRITELONG(pbuf,saalen);	/* initial length */
+    saa_rnbytes(ppubnames, pbuf, saalen);
+    saa_free(ppubnames);
+
+    /* build info section */
+    pinfo = saa_init(1L);
+    pinforel = saa_init(1L);
+    saa_write16(pinfo,2);			/* dwarf version */
+    saa_write32(pinforel, pinfo->datalen + 4);
+    saa_write32(pinforel, (dwarf_abbrevsym << 8) +  R_386_32); /* reloc to abbrev */
+    saa_write32(pinforel, 0);
+    saa_write32(pinfo,0);			/* offset into abbrev */
+    saa_write8(pinfo,4);			/* pointer size */
+    saa_write8(pinfo,1);			/* abbrviation number LEB128u */
+    saa_write32(pinforel, pinfo->datalen + 4);
+    saa_write32(pinforel, ((dwarf_fsect->section + 2) << 8) +  R_386_32);
+    saa_write32(pinforel, 0);
+    saa_write32(pinfo,0);			/* DW_AT_low_pc */
+    saa_write32(pinforel, pinfo->datalen + 4);
+    saa_write32(pinforel, ((dwarf_fsect->section + 2) << 8) +  R_386_32);
+    saa_write32(pinforel, 0);
+    saa_write32(pinfo,highaddr);		/* DW_AT_high_pc */
+    saa_write32(pinforel, pinfo->datalen + 4);
+    saa_write32(pinforel, (dwarf_linesym << 8) +  R_386_32); /* reloc to line */
+    saa_write32(pinforel, 0);
+    saa_write32(pinfo,0);			/* DW_AT_stmt_list */
+    saa_wbytes(pinfo, elf_module, strlen(elf_module)+1);
+    saa_wbytes(pinfo, nasm_signature, strlen(nasm_signature)+1);
+    saa_write16(pinfo,DW_LANG_Mips_Assembler);
+    saa_write8(pinfo,2);			/* abbrviation number LEB128u */
+    saa_write32(pinforel, pinfo->datalen + 4);
+    saa_write32(pinforel, ((dwarf_fsect->section + 2) << 8) +  R_386_32);
+    saa_write32(pinforel, 0);
+    saa_write32(pinfo,0);			/* DW_AT_low_pc */
+    saa_write32(pinfo,0);			/* DW_AT_frame_base */
+    saa_write8(pinfo,0);			/* end of entries */
+    saalen = pinfo->datalen;
+    infolen = saalen + 4;
+    infobuf = pbuf = nasm_malloc(infolen);
+    WRITELONG(pbuf,saalen);		/* initial length */
+    saa_rnbytes(pinfo, pbuf, saalen);
+    saa_free(pinfo);
+
+    /* build rela.info section */
+    inforellen = saalen = pinforel->datalen;
+    inforelbuf = pbuf = nasm_malloc(inforellen);
+    saa_rnbytes(pinforel, pbuf, saalen);
+    saa_free(pinforel); 
+
+    /* build abbrev section */
+    pabbrev = saa_init(1L);
+    saa_write8(pabbrev,1);			/* entry number LEB128u */
+    saa_write8(pabbrev,DW_TAG_compile_unit);	/* tag LEB128u */
+    saa_write8(pabbrev,1);			/* has children */
+    /* the following attributes and forms are all LEB128u values */
+    saa_write8(pabbrev,DW_AT_low_pc);
+    saa_write8(pabbrev,DW_FORM_addr);
+    saa_write8(pabbrev,DW_AT_high_pc);
+    saa_write8(pabbrev,DW_FORM_addr);
+    saa_write8(pabbrev,DW_AT_stmt_list);
+    saa_write8(pabbrev,DW_FORM_data4);
+    saa_write8(pabbrev,DW_AT_name);
+    saa_write8(pabbrev,DW_FORM_string);
+    saa_write8(pabbrev,DW_AT_producer);
+    saa_write8(pabbrev,DW_FORM_string);
+    saa_write8(pabbrev,DW_AT_language);
+    saa_write8(pabbrev,DW_FORM_data2);
+    saa_write16(pabbrev,0);			/* end of entry */
+    /* LEB128u usage same as above */
+    saa_write8(pabbrev,2);			/* entry number */
+    saa_write8(pabbrev,DW_TAG_subprogram);
+    saa_write8(pabbrev,0);			/* no children */
+    saa_write8(pabbrev,DW_AT_low_pc);
+    saa_write8(pabbrev,DW_FORM_addr);
+    saa_write8(pabbrev,DW_AT_frame_base);
+    saa_write8(pabbrev,DW_FORM_data4);
+    saa_write16(pabbrev,0);			/* end of entry */
+    abbrevlen = saalen = pabbrev->datalen;
+    abbrevbuf = pbuf = nasm_malloc(saalen);
+    saa_rnbytes(pabbrev, pbuf, saalen);
+    saa_free(pabbrev);
+
+    /* build line section */
+    /* prolog */
+    plines = saa_init(1L);
+    saa_write8(plines,1);			/* Minimum Instruction Length */
+    saa_write8(plines,1);			/* Initial value of 'is_stmt' */
+    saa_write8(plines,line_base);		/* Line Base */
+    saa_write8(plines,line_range);	/* Line Range */
+    saa_write8(plines,opcode_base);	/* Opcode Base */
+    /* standard opcode lengths (# of LEB128u operands) */
+    saa_write8(plines,0);			/* Std opcode 1 length */
+    saa_write8(plines,1);			/* Std opcode 2 length */
+    saa_write8(plines,1);			/* Std opcode 3 length */
+    saa_write8(plines,1);			/* Std opcode 4 length */
+    saa_write8(plines,1);			/* Std opcode 5 length */
+    saa_write8(plines,0);			/* Std opcode 6 length */
+    saa_write8(plines,0);			/* Std opcode 7 length */
+    saa_write8(plines,0);			/* Std opcode 8 length */
+    saa_write8(plines,1);			/* Std opcode 9 length */
+    saa_write8(plines,0);			/* Std opcode 10 length */
+    saa_write8(plines,0);			/* Std opcode 11 length */
+    saa_write8(plines,1);			/* Std opcode 12 length */
+    /* Directory Table */ 
+    saa_write8(plines,0);			/* End of table */
+    /* File Name Table */
+    ftentry = dwarf_flist;
+    for (indx = 0;indx<dwarf_numfiles;indx++)
+    {
+      saa_wbytes(plines, ftentry->filename, (int32_t)(strlen(ftentry->filename) + 1));
+      saa_write8(plines,0);			/* directory  LEB128u */
+      saa_write8(plines,0);			/* time LEB128u */
+      saa_write8(plines,0);			/* size LEB128u */
+      ftentry = ftentry->next;
+    }
+    saa_write8(plines,0);			/* End of table */
+    linepoff = plines->datalen;
+    linelen = linepoff + totlen + 10;
+    linebuf = pbuf = nasm_malloc(linelen);
+    WRITELONG(pbuf,linelen-4);		/* initial length */
+    WRITESHORT(pbuf,3);			/* dwarf version */
+    WRITELONG(pbuf,linepoff);		/* offset to line number program */
+    /* write line header */
+    saalen = linepoff;
+    saa_rnbytes(plines, pbuf, saalen);   /* read a given no. of bytes */
+    pbuf += linepoff;
+    saa_free(plines);
+    /* concatonate line program ranges */
+    linepoff += 13;
+    plinesrel = saa_init(1L);
+    psect = dwarf_fsect;
+    for (indx = 0; indx < dwarf_nsections; indx++)
+    {
+         saa_write32(plinesrel, linepoff);
+         saa_write32(plinesrel, ((uint32_t) (psect->section + 2) << 8) +  R_386_32);
+         saa_write32(plinesrel, (uint32_t) 0);
+         plinep = psect->psaa;
+         saalen = plinep->datalen;
+         saa_rnbytes(plinep, pbuf, saalen);
+         pbuf += saalen;
+         linepoff += saalen;
+         saa_free(plinep);
+         /* done with this entry */
+         psect = psect->next;
+    }
+
+
+    /* build rela.lines section */
+    linerellen =saalen = plinesrel->datalen;
+    linerelbuf = pbuf = nasm_malloc(linerellen); 
+    saa_rnbytes(plinesrel, pbuf, saalen);
+    saa_free(plinesrel);
+
+    /* build frame section */
+    framelen = 4;
+    framebuf = pbuf = nasm_malloc(framelen);
+    WRITELONG(pbuf,framelen-4);		/* initial length */
+
+    /* build loc section */
+    loclen = 16;
+    locbuf = pbuf = nasm_malloc(loclen);
+    WRITELONG(pbuf,0);		/* null  beginning offset */
+    WRITELONG(pbuf,0);		/* null  ending offset */
+}
+
+void dwarf32_cleanup(void)
+{
+    if (arangesbuf)
+        nasm_free(arangesbuf);
+    if (arangesrelbuf)
+        nasm_free(arangesrelbuf);
+    if (pubnamesbuf)
+        nasm_free(pubnamesbuf);
+    if (infobuf)
+        nasm_free(infobuf);
+    if (inforelbuf)
+        nasm_free(inforelbuf);
+    if (abbrevbuf)
+        nasm_free(abbrevbuf);
+    if (linebuf)
+        nasm_free(linebuf);
+    if (linerelbuf)
+        nasm_free(linerelbuf);
+    if (framebuf)
+        nasm_free(framebuf);
+    if (locbuf)
+        nasm_free(locbuf);
+}
+void dwarf32_findfile(const char * fname)
+{
+   int finx;
+   struct linelist *match;
+
+   /* return if fname is current file name */
+   if (dwarf_clist && !(strcmp(fname, dwarf_clist->filename))) return;
+   /* search for match */
+   else 
+   {
+     match = 0;
+     if (dwarf_flist)
+     {
+       match = dwarf_flist;
+       for (finx = 0; finx < dwarf_numfiles; finx++)
+       {
+         if (!(strcmp(fname, match->filename)))
+         {
+	   dwarf_clist = match;
+           return;
+         }
+       }
+     }
+     /* add file name to end of list */
+     dwarf_clist =  (struct linelist *)nasm_malloc(sizeof(struct linelist));
+     dwarf_numfiles++;
+     dwarf_clist->line = dwarf_numfiles;
+     dwarf_clist->filename = nasm_malloc(strlen(fname) + 1);
+     strcpy(dwarf_clist->filename,fname);
+     dwarf_clist->next = 0;
+     /* if first entry */
+     if (!dwarf_flist)
+     {
+       dwarf_flist = dwarf_elist = dwarf_clist;
+       dwarf_clist->last = 0;
+     }
+     /* chain to previous entry */
+     else
+     {
+       dwarf_elist->next = dwarf_clist;
+       dwarf_elist = dwarf_clist;
+     }
+   }
+}
+/*  */
+void dwarf32_findsect(const int index)
+{
+   int sinx;
+   struct sectlist *match;
+   struct SAA *plinep;
+   /* return if index is current section index */
+   if (dwarf_csect && (dwarf_csect->section == index))
+   {
+      return;
+   }
+   /* search for match */
+   else 
+   {
+     match = 0;
+     if (dwarf_fsect)
+     {
+       match = dwarf_fsect;
+       for (sinx = 0; sinx < dwarf_nsections; sinx++)
+       {
+         if ((match->section == index))
+         {
+	   dwarf_csect = match;
+           return;
+         }
+        match = match->next;
+       }
+     }
+     /* add entry to end of list */
+     dwarf_csect =  (struct sectlist *)nasm_malloc(sizeof(struct sectlist));
+     dwarf_nsections++;
+     dwarf_csect->psaa = plinep = saa_init(1L);
+     dwarf_csect->line = 1;
+     dwarf_csect->offset = 0;
+     dwarf_csect->file = 1;
+     dwarf_csect->section = index;
+     dwarf_csect->next = 0;
+     /* set relocatable address at start of line program */
+     saa_write8(plinep,DW_LNS_extended_op);
+     saa_write8(plinep,5);			/* operand length */
+     saa_write8(plinep,DW_LNE_set_address);
+     saa_write32(plinep,0);		/* Start Address */
+     /* if first entry */
+     if (!dwarf_fsect)
+     {
+       dwarf_fsect = dwarf_esect = dwarf_csect;
+       dwarf_csect->last = 0;
+     }
+     /* chain to previous entry */
+     else
+     {
+       dwarf_esect->next = dwarf_csect;
+       dwarf_esect = dwarf_csect;
+     }
+   }
+}
 
 #endif                          /* OF_ELF */
